<?php

namespace Drupal\biko_field_values\Repository;

use Drupal\biko_field_values\WrappedEntity;
use Drupal\Core\Datetime\DrupalDateTime;
use Drupal\Core\Entity\EntityInterface;

abstract class WrappedEntityQuery {

  protected const DEFAULT_DATE_INTERVAL_FIELD_NAME = NULL;

  abstract public function getEntityTypeId(): string;

  abstract protected function getEntityIdFieldName(): string;

  abstract public function getBundle(): string;

  protected string $conjunction;
  protected array $conditions = [];
  protected array $sorts = [];
  protected int $start = 0;
  protected ?int $length = NULL;
  protected array $entityRelations = [];
  public function __construct(string $conjunction = 'AND') {
    $this->conjunction = $conjunction;
  }

  /**
   * more convenient constructor to avoid (new WrappedEntityQuery())-> and use WrappedEntityQuery::new()->
   *
   * @param string $conjunction
   *
   * @return static
   */
  public static function new(string $conjunction = 'AND'): self {
    return new static($conjunction);
  }

  public function getConditions(): array {
    return $this->conditions;
  }

  public function getSorts(): array {
    return $this->sorts;
  }

  public function getConjunction(): string {
    return $this->conjunction;
  }

  /**
   * @return \Drupal\biko_field_values\Repository\WrappedEntityQueryRelation[]
   */
  public function getEntityRelations(): array {
    return $this->entityRelations;
  }

  /**
   * @param string|\Drupal\biko_field_values\Repository\WrappedEntityQuery $fieldOrQuery
   * @param mixed|null $value
   * @param string|null $operator
   * @param string|null $langcode
   *
   * @return self
   */
  public function condition($fieldOrQuery, $value = NULL, string $operator = NULL, string $langcode = NULL): self {
    $this->conditions[] = [
      'field' => $fieldOrQuery,
      'value' => $value,
      'operator' => $operator,
      'langcode' => $langcode,
    ];
    return $this;
  }

  /**
   * @param string $direction
   * @param string|null $fieldName
   *
   * @return self
   */
  public function sort(string $direction = 'ASC', ?string $fieldName = NULL): self {

    if (!in_array($direction, ['ASC', 'DESC'], TRUE)) {
      throw new \InvalidArgumentException(sprintf("'%s' is not a valid sorting direction", $direction));
    }

    $fieldName = $fieldName ?? static::DEFAULT_DATE_INTERVAL_FIELD_NAME;
    if (!$fieldName) {
      throw new \InvalidArgumentException(sprintf("Field name argument has not been set and there is not default date interval field name for '%s' class", get_class($this)));
    }

    $this->sorts[] = [
      'field' => $fieldName,
      'direction' => $direction,
    ];

    return $this;

  }

  /**
   * @param string|null $start
   * @param string|null $length
   *
   * @see \Drupal\Core\Entity\Query\QueryInterface::range()
   *
   * @return self
   */
  public function range(?int $start = NULL, ?int $length = NULL): self {

    $this->start = $start ?? 0;
    $this->length = $length;

    return $this;

  }

  public function getStart(): ?int {
    return $this->start;
  }

  public function getLength(): ?int {
    return $this->length;
  }

  /**
   * @param \Drupal\biko_field_values\WrappedEntity[] $wrappedEntitys
   *
   * @return self
   */
  public function in(array $wrappedEntitys): self {
    return $this->idIn(
      array_map(
        static function (WrappedEntity $wrappedEntity) {
          return $wrappedEntity->getEntity()->id();
        }, $wrappedEntitys
      )
    );
  }

  /**
   * @param int[] $entityIds
   *
   * @return self
   */
  public function idIn(array $entityIds): self {
    $this->condition(
      $this->getEntityIdFieldName(),
      $entityIds,
      'IN'
    );
    return $this;
  }

  /**
   * @param \Drupal\biko_field_values\WrappedEntity[] $wrappedEntitys
   *
   * @return self
   */
  public function notIn(array $wrappedEntitys): self {
    return $this->idNotIn(
      array_map(
        static function (WrappedEntity $wrappedEntity) {
          return $wrappedEntity->getEntity()->id();
        }, $wrappedEntitys
      )
    );
  }

  /**
   * @param int[] $entityIds
   *
   * @return self
   */
  public function idNotIn(array $entityIds): self {
    $this->condition(
      $this->getEntityIdFieldName(),
      $entityIds,
      'NOT IN'
    );
    return $this;
  }

  /**
   * @param \Drupal\biko_field_values\WrappedEntity $wrappedEntity
   *
   * @return self
   */
  public function equals(WrappedEntity $wrappedEntity): self {
    $this->in([$wrappedEntity]);
    return $this;
  }

  /**
   * @param \Drupal\biko_field_values\WrappedEntity $wrappedEntity
   *
   * @return self
   */
  public function notEquals(WrappedEntity $wrappedEntity): self {
    $this->notIn([$wrappedEntity]);
    return $this;
  }

  /**
   * @param \Drupal\biko_field_values\WrappedEntity $wrappedEntity
   * @param string|null $fieldPath
   *
   * @return self
   */
  public function relatedWith(WrappedEntity $wrappedEntity, ?string $fieldPath = NULL): self {
    return $this->relatedWithEntity($wrappedEntity->getEntity(), $fieldPath);
  }

  /**
   * This method must be used directly ONLY for entities not having an entity service
   * (i.e. some less often used vocabularies) For entities with an WrappedEntity,
   * the method ::relatedWith() must be used.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   * @param string|null $fieldPath
   *
   * @return self
   */
  public function relatedWithEntity(EntityInterface $entity, ?string $fieldPath = NULL): self {

    $entityRelation = new WrappedEntityQueryRelation();
    $entityRelation->entity = $entity;
    $entityRelation->fieldPath = $fieldPath;
    $this->entityRelations[] = $entityRelation;

    return $this;

  }

  protected const DATE_INTERVAL_CONDITION_PAST = 'past';
  protected const DATE_INTERVAL_CONDITION_CURRENT = 'current';
  protected const DATE_INTERVAL_CONDITION_FUTURE = 'future';

  /**
   * @param string|null $dateIntervalFieldName
   *
   * @return self
   */
  public function current(?string $dateIntervalFieldName = NULL): self {
    return $this->dateIntervalCondition(self::DATE_INTERVAL_CONDITION_CURRENT, $dateIntervalFieldName);
  }

  /**
   * @param string|null $dateIntervalFieldName
   *
   * @return self
   */
  public function past(?string $dateIntervalFieldName = NULL): self {
    return $this->dateIntervalCondition(self::DATE_INTERVAL_CONDITION_PAST, $dateIntervalFieldName);
  }

  /**
   * @param string|null $dateIntervalFieldName
   *
   * @return self
   */
  public function future(?string $dateIntervalFieldName = NULL): self {
    return $this->dateIntervalCondition(self::DATE_INTERVAL_CONDITION_FUTURE, $dateIntervalFieldName);
  }

  /**
   * @param string $condition
   * @param string|null $dateIntervalFieldName
   *
   * @return self
   */
  protected function dateIntervalCondition(string $condition, ?string $dateIntervalFieldName = NULL): self {

    $dateIntervalFieldName = $dateIntervalFieldName ?? static::DEFAULT_DATE_INTERVAL_FIELD_NAME;

    $today = new DrupalDateTime();
    $startOfToday = (clone $today)->setTime(0, 0)->format('Y-m-d\TH:i:s', ['timezone' => 'UTC']);
    $endOfToday = (clone $today)->setTime(23, 59, 59)->format('Y-m-d\TH:i:s', ['timezone' => 'UTC']);

    $startFieldName = $dateIntervalFieldName . '.value';
    $endFieldName = $dateIntervalFieldName . '.end_value';

    switch ($condition) {

      case self::DATE_INTERVAL_CONDITION_PAST:
        $this->condition($endFieldName, $startOfToday, '<');
        break;

      case self::DATE_INTERVAL_CONDITION_CURRENT:
        $this->condition(
          (new static('AND'))
            ->condition($startFieldName, $endOfToday, '<=')
            ->condition($endFieldName, $startOfToday, '>=')
        );
        break;

      case self::DATE_INTERVAL_CONDITION_FUTURE:
        $this->condition($startFieldName, $endOfToday, '>');
        break;

      default:
        throw new \InvalidArgumentException(sprintf("Invalid condition '%s'", $condition));

    }

    return $this;

  }

  /**
   * @param \Drupal\Core\Datetime\DrupalDateTime $date
   *
   * @return self
   */
  public function fromDate(DrupalDateTime $date): self {
    return $this->betweenDates($date, NULL);
  }

  /**
   * @param \Drupal\Core\Datetime\DrupalDateTime $date
   *
   * @return self
   */
  public function toDate(DrupalDateTime $date): self {
    return $this->betweenDates(NULL, $date);
  }

  /**
   * @param \Drupal\Core\Datetime\DrupalDateTime|null $startDate
   * @param \Drupal\Core\Datetime\DrupalDateTime|null $endDate
   *
   * @return self
   */
  public function betweenDates(?DrupalDateTime $startDate = NULL, ?DrupalDateTime $endDate = NULL): self {

    if ($startDate) {
      $this->condition(static::DEFAULT_DATE_INTERVAL_FIELD_NAME . '.end_value', $startDate->format('Y-m-d') . 'T00:00:00', '>=');
    }
    if ($endDate) {
      $this->condition(static::DEFAULT_DATE_INTERVAL_FIELD_NAME . '.value', $endDate->format('Y-m-d') . 'T23:59:59', '<=');
    }

    return $this;

  }

}
