<?php

namespace Drupal\biko_field_values\Repository;

use Drupal\biko_field_values\WrappedEntityFactory;
use Drupal\Core\Entity\EntityFieldManagerInterface;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Entity\Query\ConditionInterface;
use Drupal\Core\Entity\Query\QueryInterface;

/**
 * Represents an entity repository.
 *
 * Different entities must inherit this class
 */
abstract class WrappedEntityRepository {

  public const ENTITY_KEY__BUNDLE = NULL;

  /**
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entityTypeManager
   * @param \Drupal\biko_field_values\WrappedEntityFactory $wrappedEntityFactory
   * @param \Drupal\Core\Entity\EntityFieldManagerInterface $entityFieldManager
   */
  public function __construct(
    protected EntityTypeManagerInterface $entityTypeManager,
    protected WrappedEntityFactory $wrappedEntityFactory,
    protected EntityFieldManagerInterface $entityFieldManager
  ) {}

  /**
   * Processes an array of WrappedEntityQuery conditions into a query condition group,
   * ready for \Drupal\Core\Entity\Query\QueryInterface::condition().
   *
   * @param array $condition
   *   An array of condition parameters, same as in QueryInterface::condition():
   *     - field
   *     - value
   *     - operator
   *     - langcode
   *
   *     Except that 'field' parameter could be a field name or another WrappedEntityQuery,
   *     allowing condition nesting.
   *
   * @param \Drupal\Core\Entity\Query\QueryInterface $query
   *   An entity query is neccessary to construct the or/and condition groups.
   *
   * @return \Drupal\Core\Entity\Query\ConditionInterface
   */
  protected function processCondition(array $condition, QueryInterface $query): ConditionInterface {

    $fieldNameOrQuery = $condition['field'];

    if (is_string($fieldNameOrQuery)) {
      return $query->andConditionGroup()->condition(
        $condition['field'],
        $condition['value'],
        $condition['operator'],
        $condition['langcode'],
      );
    }

    if ($fieldNameOrQuery instanceof WrappedEntityQuery) {
      $conditionGroup =
        $fieldNameOrQuery->getConjunction() === 'AND'
        ? $query->andConditionGroup()
        : $query->orConditionGroup();
      foreach ($fieldNameOrQuery->getConditions() as $subCondition) {
        $conditionGroup->condition($this->processCondition($subCondition, $query));
      }
      return $conditionGroup;
    }

    throw new \InvalidArgumentException("\$condition['field'] must be a string or an WrappedEntityQuery");

  }

  /**
   * Prepares the Entity query from an WrappedEntityQuery.
   *
   * Isolating this method from ::get() allows child classes altering the query.
   *
   * @param \Drupal\biko_field_values\Repository\WrappedEntityQuery $wrappedEntityQuery
   *
   * @return \Drupal\Core\Entity\Query\QueryInterface
   */
  public function getQuery(WrappedEntityQuery $wrappedEntityQuery): QueryInterface {

    $entityStorage = $this->entityTypeManager->getStorage($wrappedEntityQuery->getEntityTypeId());
    $query = $entityStorage->getQuery($wrappedEntityQuery->getConjunction());
    $query->accessCheck(TRUE);

    // Bundle condition.
    $query->condition(static::ENTITY_KEY__BUNDLE, $wrappedEntityQuery->getBundle(), '=');

    // Query conditions.
    foreach ($wrappedEntityQuery->getConditions() as $condition) {
      $query->condition($this->processCondition($condition, $query));
    }

    // Query entity relations.
    foreach ($wrappedEntityQuery->getEntityRelations() as $entityRelation) {

      $fieldPath = $entityRelation->fieldPath;

      // Try to find the field, if not specified.
      if (NULL === $fieldPath) {
        $candidateFields = $this->getCandidateEntityReferenceFieldNamesReferencingToEntity($wrappedEntityQuery, $entityRelation->entity);
        if (count($candidateFields) === 0) {
          throw new \Exception(
            sprintf(
              "There are no entity reference fields that could relate this entity to a '%s:%s', so you must specify the field path",
              $wrappedEntityQuery->getEntityTypeId(),
              $wrappedEntityQuery->getBundle()
            )
          );
        }
        if (count($candidateFields) > 1) {
          throw new \Exception(
            sprintf(
              "There are multiple entity reference fields that could relate this entity to a '%s:%s', so you must specify the field path",
              $wrappedEntityQuery->getEntityTypeId(),
              $wrappedEntityQuery->getBundle()
            )
          );
        }
        $fieldPath = reset($candidateFields);
      }

      $query->condition($fieldPath, $entityRelation->entity->id(), '=');

    }

    // Sorting.
    foreach ($wrappedEntityQuery->getSorts() as $sort) {
      $query->sort(
        $sort['field'],
        $sort['direction']
      );
    }

    // Paging.
    if ($wrappedEntityQuery->getStart() === 0 && $wrappedEntityQuery->getLength() === NULL) {
      $query->range(NULL);
    }
    else {
      $query->range($wrappedEntityQuery->getStart(), $wrappedEntityQuery->getLength());
    }

    return $query;

  }

  /**
   * Returns entity services according to supplied query object.
   *
   * @param \Drupal\biko_field_values\Repository\WrappedEntityQuery $wrappedEntityQuery
   *
   * @return \Drupal\biko_field_values\WrappedEntity[]
   */
  public function get(WrappedEntityQuery $wrappedEntityQuery): array {

    $query = $this->getQuery($wrappedEntityQuery);

    $entityStorage = $this->entityTypeManager->getStorage($wrappedEntityQuery->getEntityTypeId());

    $wrappedEntitys = [];
    foreach ($entityStorage->loadMultiple($query->execute()) as $entity) {
      $wrappedEntitys[] = $this->wrappedEntityFactory->wrapOrFail($entity);
    }

    return $wrappedEntitys;

  }

  /**
   * Returns total count without ranging the query.
   *
   * @param \Drupal\biko_field_values\Repository\WrappedEntityQuery $wrappedEntityQuery
   *
   * @return int
   */
  public function getTotalCount(WrappedEntityQuery $wrappedEntityQuery): int {

    return $this->getQuery($wrappedEntityQuery)
      ->range()
      ->count()
      ->execute();

  }

  /**
   * @param \Drupal\biko_field_values\Repository\WrappedEntityQuery $wrappedEntityQuery
   * @param \Drupal\taxonomy\TermInterface $targetEntity
   *
   * @return string[]
   */
  protected function getCandidateEntityReferenceFieldNamesReferencingToEntity(WrappedEntityQuery $wrappedEntityQuery, EntityInterface $targetEntity): array {

    /** @var \Drupal\Core\Entity\EntityFieldManagerInterface $entityFieldManager */
    $entityFieldManager = \Drupal::service('entity_field.manager');

    $candidateFields = [];
    foreach ($entityFieldManager->getFieldDefinitions($wrappedEntityQuery->getEntityTypeId(), $wrappedEntityQuery->getBundle()) as $fieldDefinition) {
      $fieldDefinitionSettings = $fieldDefinition->getSettings();
      if (
        $fieldDefinition->getType() === 'entity_reference'
        && $fieldDefinitionSettings['target_type'] === $targetEntity->getEntityTypeId()
        && in_array($targetEntity->bundle(), $fieldDefinitionSettings['handler_settings']['target_bundles'] ?? [], TRUE)
      ) {
        $candidateFields[] = $fieldDefinition->getName();
      }
    }

    return $candidateFields;

  }

}
